/*
 * sound.v
 *
 * Circuit to generate PDM sound using an Altera Nios board with an
 * AleaRep Lancelot sound add-on.
 *
 * This module should be fed with a 100 MHz clock from the PLL.
 * 
 * This is a second-order delta-sigma DAC, assuming an external
 * low pass filter.
 */

//
// The number of extra bits of precision (XBITS) necessary isn't very
// clear.  I ran the contents of a full CD through a simulator of the
// algorithm, and came up with at least one transition which required
// 6 extra bits.
//
`define MSB	(BITS-1)
`define XBITS   6		// Additional bits of precision
`define DBITS	(`MSB+`XBITS)

module sound(
	     reset_n,
	     clk,
	     data,
	     audio,
	     audio_clk,
	     );

   parameter BITS = 16;

   input          reset_n;
   input	  clk;
   input [`MSB:0] data;
   output	  audio;
   output         audio_clk;

   reg [`MSB:0] 	data_latch;
   wire [`DBITS:0] 	xdata;	// Expanded data

   wire [`DBITS:0] 	delta2_add;
   wire [`DBITS:0] 	delta_add;

   wire [`DBITS:0] 	sigma2_add;
   wire [`DBITS:0] 	sigma_add;

   reg [`DBITS:0] 	sigma2_latch;
   reg [`DBITS:0] 	sigma_latch;

   wire [`DBITS:0] 	delta;

   reg [11:0] 		pulse_ctr = 0;
   
   // Note: the input is expected *signed*.  For unsigned input, invert
   // bit `MSB.  This is a change from previous code.
   assign	xdata[`DBITS:BITS] = {`XBITS{data_latch[`MSB]}};
   assign	xdata[`MSB:0] = data_latch;

   assign	delta2_add = xdata + delta;
   assign	sigma2_add = delta2_add + sigma2_latch;
   
   assign 	delta_add = sigma2_add + delta;
   assign 	sigma_add = delta_add + sigma_latch;

   // delta is -32768 if sigma_latch is negative, otherwise 32768
   assign 	delta[`DBITS:BITS] = {`XBITS{sigma_latch[`DBITS]}};
   assign	delta[`MSB] = 1'b1;
   assign 	delta[`MSB-1:0] = 0;

   always @(negedge reset_n or posedge clk)
     if (~reset_n)
       begin
	  pulse_ctr <= 0;
	  sigma_latch <= 0;
	  sigma2_latch <= 0;
       end
        else
       begin
	  if ( pulse_ctr == 0 )
	    begin
	       // 2268 is the oversampling ratio (100 MHz/44100 Hz)
	       pulse_ctr <= 2268-1;
	       data_latch <= data;
	    end
	  else
	    begin
	       pulse_ctr <= pulse_ctr-1;
	    end

	  sigma_latch  <= sigma_add;
	  sigma2_latch <= sigma2_add;
       end

   assign audio     = sigma_latch[`DBITS];
   assign audio_clk = pulse_ctr[11];
   
endmodule // sound
